import pandas as pd
import numpy as np
from exclusiveAI.utils import train_split

__all__ = [
    "read_cup_training_dataset",
    "train_val_test_split",
    "read_cup_test_dataset"
]


def read_cup_training_dataset(path="exclusiveAI/datasets"):
    training_set_filepath = f"{path}/ML-CUP23-TR.csv"

    training_set_df = pd.read_csv(
        training_set_filepath,
        sep=",",
        dtype='float64',
        names=["id", "f1", "f2", "f3", "f4", "f5", "f6", "f7", "f8", "f9", "f10", "TargetX", "TargetY", "TargetZ"],
        skiprows=7,
    ).set_index("id")

    # Extract and remove the last three columns as labels
    training_set_labels = np.array(training_set_df[["TargetX", "TargetY", "TargetZ"]].copy())
    training_set_df.drop(["TargetX", "TargetY", "TargetZ"], axis=1, inplace=True)

    training_set_df = np.array(training_set_df)

    return training_set_df, training_set_labels


def train_val_test_split(training_set_df, training_set_labels, train_size=0.7, test_size=0.33):
    train_set, train_labels, tmp_set, tmp_labels, train_idx, tmp_idx = train_split(training_set_df,
                                                                                   training_set_labels,
                                                                                   split_size=1 - train_size)
    val_set, val_labels, test_set, test_labels, val_idx, test_idx = train_split(tmp_set,
                                                                                tmp_labels,
                                                                                split_size=test_size)

    return (train_set, train_labels,
            val_set, val_labels, test_set,
            test_labels, train_idx,
            tmp_idx[val_idx], tmp_idx[test_idx])


def read_cup_test_dataset(path='exclusiveAI/datasets'):
    test_set_filepath = f"{path}/ML-CUP23-TS.csv"

    test_set_df = pd.read_csv(
        test_set_filepath,
        sep=",",
        dtype='float64',
        names=["id", "f1", "f2", "f3", "f4", "f5", "f6", "f7", "f8", "f9", "f10"],
        skiprows=7,
    ).set_index("id")

    test_set_df = np.array(test_set_df)

    return test_set_df



from .Initializer import Initializer
import numpy as np

__all__ = ["Uniform"]


class Uniform(Initializer):
    """
    Initialize weights from a random uniform distribution.
    Args:
        low (float): Lower bound of the uniform distribution.
        high (float): Upper bound of the uniform distribution.
    Attributes:
        name (str): Name of the initializer.
        low (float): Lower bound of the uniform distribution.
        high (float): Upper bound of the uniform distribution.
    """
    def __init__(self, low=-0.5, high=0.5):
        super().__init__(name='Uniform')
        self.low = low
        self.high = high

    def initialize(self, shape):
        """
        Args:
            shape (tuple): shape of the weights vector.
        Returns:
            object: returns an array of shape `shape` with values drawn from a random uniform distribution.
        """
        return np.random.uniform(self.low, self.high, shape)



from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
import numpy as np
from joblib import Parallel, delayed
from exclusiveAI.Composer import Composer
from .ValidationUtils import get_best_model
from tqdm import tqdm
from functools import partial

__all__ = ["validate", "double_validate"]

"""
K-fold cross validation.
"""


def split(x, y_true=None, random_state=None, shuffle=True, n_splits=5):
    """
    Data splitting for cross validation
    Args:
        x: input
        y_true: target
        random_state: random seed
        shuffle: shuffle
        n_splits: number of splits for cross validation

    Raises:
        ValueError: if y_true is not specified and shuffle is True.

    Returns: a tuple containing two lists of indices:
            - The training indices obtained by excluding the validation part in the fold.
            - The validation indices for the current fold.
    """
    if y_true is None and shuffle:
        raise ValueError("y must be specified if shuffle is True")

    indices = np.arange(len(x))  # create an array of indices
    if shuffle:
        np.random.seed(random_state)
        np.random.shuffle(indices)

    # Calculate the base size of each fold.
    fold_sizes = np.full(n_splits, len(x) // n_splits, dtype=int)
    # The remainder of this division is distributed to ensure that all samples are included.
    fold_sizes[:len(x) % n_splits] += 1

    current = 0
    for fold_size in fold_sizes:
        start, stop = current, current + fold_size
        first_part = list(indices[np.concatenate((indices[:start], indices[stop:]))])
        last_part = list(indices[start:stop])
        yield first_part, last_part
        current = stop


def validate_single_config(config, x, y_true, epochs, batch_size, disable_line, random_state, shuffle, n_splits,
                           metric, regression, number_of_initializations):
    best = None
    score = 0
    for train_index, val_index in split(x, y_true, random_state, shuffle, n_splits):
        x_train, x_val = x[train_index], x[val_index]
        y_train, y_val = y_true[train_index], y_true[val_index]

        internal_score = 0
        model = None

        for i in range(number_of_initializations):
            tmp_model = Composer(config=config).compose(regression)
            tmp_model.train(inputs=x_train, input_label=y_train, val=x_val, val_labels=y_val,
                            disable_line=disable_line, epochs=epochs, batch_size=batch_size)
            if model is None or np.min(model.history[metric]) > np.min(tmp_model.history[metric]):
                model = tmp_model
            internal_score += np.min(tmp_model.history[metric])
        internal_score /= number_of_initializations

        if best is None or best.get_last()[metric] < score:
            best = model
            score = internal_score
    return score, best, config


def validate(configs, x, y_true, metric='val_mse', max_configs=1, random_state=None, shuffle=True, parallel=True,
             n_splits=5, number_of_initializations=1, epochs=100, assessment=False, batch_size=32,
             disable_line=True, workers=None, regression=False, return_models_history=False):
    evaluate_partial = partial(validate_single_config, x=x.copy(), y_true=y_true.copy(), epochs=epochs,
                               batch_size=batch_size, disable_line=disable_line, random_state=random_state,
                               shuffle=shuffle, n_splits=n_splits, metric=metric, regression=regression,
                               number_of_initializations=number_of_initializations)

    evaluator = ProcessPoolExecutor(max_workers=workers) if parallel else None
    results_func = evaluator.map if evaluator else map

    results = list(tqdm(results_func(evaluate_partial, configs), total=len(configs), desc="Models", colour="white"))

    # results = Parallel(n_jobs=workers)(
    #     delayed(validate_single_config)(config, x.copy(), y_true.copy(), epochs, batch_size, disable_line,
    #                                     random_state, shuffle, n_splits, metric, regression, number_of_initializations)
    #     for config in
    #     tqdm(configs, desc="Models", colour="white")
    # )

    best_models, best_configs = get_best_model(results, max_configs)

    if return_models_history:
        if max_configs > 1:
            return [model.history for model in best_models], best_configs if not assessment else [
                model.get_last()[metric] for model in best_models]
        return best_models[0].history, best_configs[0] if not assessment else best_models[0].get_last()[metric]

    if max_configs > 1:
        return best_configs if not assessment else [model.get_last()[metric] for model in best_models]
    return best_configs[0] if not assessment else best_models[0].get_last()[metric]


def double_validate(configs, x, y_true, metric='val_mse', inner_splits=4, random_state=None,
                    shuffle=True, n_splits=5, number_of_initializations=1,
                    epochs=100, regression=False, return_models_history=False,
                    batch_size=32, disable_line=True, workers=-1):
    """
    Perform a double k-fold cross validation
    Args:
        configs: A dictionary containing configuration parameters for the models
        x: the input
        y_true: the target
        metric: the metric to be used for the cross validation
        inner_splits: the number of splits for the cross validation
        shuffle (bool): true if you want to shuffle the data
        n_splits (int): Number of splits
        number_of_initializations (int): Number of initializations for each model
        random_state (int): random seed
        epochs (int): number of epochs to train the model
        regression (bool): true if you want to use a regression model
        batch_size (int): batch size
        disable_line (bool): whether to disable line
        eps (float): epsilon for model dimension selection
        workers (int): number of workers to use for parallel

    Returns: the best configuration found

    """

    outer_scores = []
    for outer_train_index, outer_val_index in split(x, y_true, shuffle=shuffle, n_splits=n_splits,
                                                    random_state=random_state):
        x_train, x_test = x[outer_train_index], x[outer_val_index]
        y_train, y_test = y_true[outer_train_index], y_true[outer_val_index]

        my_config = validate(configs, x=x_train, y_true=y_train, metric=metric, random_state=random_state,
                             shuffle=False, number_of_initializations=number_of_initializations,
                             n_splits=inner_splits, epochs=epochs, batch_size=batch_size, disable_line=disable_line,
                             regression=regression, workers=workers)
        model = Composer(config=my_config).compose(regression)
        model.train(x_train, y_train, epochs=epochs, batch_size=batch_size, disable_line=disable_line)
        # test error for each external split
        if return_models_history:
            outer_scores.append((model.history, model.evaluate(x_test, y_test)[0]))
        else:
            outer_scores.append(model.evaluate(x_test, y_test)[0])

    return np.mean(outer_scores), outer_scores



from itertools import product, permutations
from tqdm import tqdm
import numpy as np
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from functools import partial


class ConfiguratorGen:
    """
    This class is used to generate configurations for the Composer class.

    Args:
        random (bool, optional): If True, generates random configurations. Defaults to False.
        max_configs (int, optional): Number of configurations to generate. Only used if random is True. Defaults to 100.
        output_activation (str, optional): Activation function for the output layer. Defaults to 'linear'.
        regularizations (list, optional): List of regularizations to be used. Defaults to None.
        learning_rates (list, optional): List of learning rates to be used. Defaults to None.
        loss_function (str, optional): Loss function to be used. Default to linear.
        activation_functions (list, optional): List of activation functions to be used. Defaults to None.
        number_of_units (list, optional): List of number of units to be used. Defaults to None.
        number_of_layers (list, optional): List of number of layers to be used. Defaults to None.
        momentums (list, optional): List of momentums to be used. Defaults to None.
        optimizer (list, optional): List of optimizers to be used. Defaults to None.
        initializers (list, optional): List of initializers to be used. Defaults to None.
        input_shapes (tuple): List of input shapes to be used. Defaults to None.
        callbacks (list, optional): List of callbacks to be used. Defaults to None.
        verbose (bool, optional): Whether to print out the progress of the model. Defaults to False.
        outputs (int, optional): Number of outputs to be used. Defaults to 1.
    """

    def __init__(self,
                 output_activation: str | list,
                 loss_function: str | list,
                 optimizer: str | list,
                 activation_functions: list,
                 input_shapes: int | tuple,
                 number_of_layers: list,
                 number_of_units: list,
                 learning_rates: list,
                 initializers: list,
                 callbacks: list,
                 number_of_initializations=1,
                 regularizations=None,
                 nesterov=False,
                 momentums=None,
                 show_line=False,
                 max_configs=100,
                 verbose=False,
                 random=False,
                 beta2=None,
                 outputs=1
                 ):
        """

        """
        if regularizations is None:
            regularizations = [0]
        if optimizer is None or optimizer == []:
            optimizer = ['sgd']
        if momentums is None:
            momentums = [0]
        self.output_activation = output_activation \
            if isinstance(output_activation, list) else output_activation
        self.loss_function = loss_function \
            if isinstance(loss_function, list) else [loss_function]
        self.optimizer = optimizer \
            if isinstance(optimizer, list) else [optimizer]

        self.activation_functions = activation_functions
        self.number_of_layers = number_of_layers
        self.number_of_units = number_of_units
        self.regularizations = regularizations
        self.learning_rates = learning_rates
        self.initializers = initializers
        self.input_shapes = input_shapes
        self.callbacks = callbacks
        self.nesterov = ["True", "False"] if nesterov else ["False"]
        self.verbose = verbose
        self.outputs = outputs
        self.beta2 = beta2

        self.type = 'random' if random else 'grid'
        self.num_of_configurations = max_configs

        if beta2:
            configurations = product(regularizations,
                                     learning_rates,
                                     loss_function,
                                     momentums,
                                     optimizer,
                                     number_of_layers,
                                     initializers,
                                     self.nesterov,
                                     beta2
                                     )
        else:
            configurations = product(regularizations,
                                     learning_rates,
                                     loss_function,
                                     momentums,
                                     optimizer,
                                     number_of_layers,
                                     initializers,
                                     self.nesterov,
                                     )

        selected_configs = list(configurations)

        final_configs = self.par_combinations(selected_configs, self.activation_functions, self.number_of_units,
                                              self.output_activation, self.input_shapes, self.callbacks, self.verbose,
                                              self.outputs)
        # with tqdm(total=len(selected_configs) * number_of_initializations, desc="2nd for", colour="white",
        #           disable=not show_line) as pbar:
        #     final_configs = []
        #     for config in selected_configs:
        #         internal_config = list(config)
        #         num_layers = internal_config[5]
        #         product_unit_activation = self.units_activations_per_layer_combinations(self.activation_functions,
        #                                                                                 self.number_of_units,
        #                                                                                 num_layers)
        #         for unit, activation in product_unit_activation:
        #             local_config = internal_config[:6] + [unit, activation] + internal_config[6:]
        #             final_configs.append(local_config)
        #         pbar.update(1)

        if number_of_initializations > 1:
            with tqdm(total=len(final_configs) * number_of_initializations, desc="1st for", colour="white",
                      disable=not show_line) as pbar:
                tmp_configurations = []
                for config in final_configs:
                    for i in range(number_of_initializations):
                        tmp_configurations.append(config)
                        pbar.update(1)

                final_configs = tmp_configurations

        if self.type == 'random':
            indices = np.random.permutation(len(final_configs))
            indices = indices[:self.num_of_configurations] if indices.size > self.num_of_configurations else indices
            final_configs = [final_configs[i] for i in indices]
        self.configs = list(final_configs)
        self.current = -1
        self.max = max(max_configs, len(self.configs)) if self.type == 'random' else len(self.configs)

    @staticmethod
    def add_units_activations(config, activation_functions, number_of_units, pbar):
        final_configs = []
        internal_config = list(config)
        num_layers = internal_config[5]
        product_unit_activation = ConfiguratorGen.units_activations_per_layer_combinations(activation_functions,
                                                                                           number_of_units,
                                                                                           num_layers)
        for unit, activation in product_unit_activation:
            local_config = internal_config[:6] + [unit, activation] + internal_config[6:]
            final_configs.append(local_config)
        pbar.update(1)
        return final_configs

    @staticmethod
    def par_combinations(selected_configs, activation_functions, number_of_units,
                         output_activation, input_shapes, callbacks, verbose, outputs):

        batch_size = 1000  # You can adjust the batch size based on your requirements
        with tqdm(total=len(selected_configs), desc="Appending") as pbar:
            with ThreadPoolExecutor(max_workers=8) as executor:
                batches = [selected_configs[i:i + batch_size] for i in range(0, len(selected_configs), batch_size)]

                batch_results = []
                for batch in batches:
                    evaluate_partial_batch = partial(ConfiguratorGen.add_units_activations,
                                                     activation_functions=activation_functions,
                                                     number_of_units=number_of_units, pbar=pbar)
                    batch_results.append(executor.map(evaluate_partial_batch, batch))

        output = []
        counter = 0
        final_configs = []
        for batch in batch_results:
            for element in batch:
                for config in element:
                    final_configs.append(config)
        with tqdm(total=len(final_configs), desc="Appending") as pbar:
            for config in final_configs:
                config = {"regularization": config[0], "learning_rate": config[1], "loss_function": config[2],
                          "activation_functions": list(config[7]), "output_activation": output_activation,
                          "num_of_units": list(config[6]), "num_layers": config[5], "momentum": config[3],
                          "optimizer": config[4],
                          "initializers": config[8], "nesterov": True if config[9] == 'True' else False,
                          "beta2": config[10] if len(config) > 10 else 0,
                          "input_shape": input_shapes, "callbacks": callbacks, "verbose": verbose,
                          "outputs": outputs, "model_name": 'Model' + str(counter)}
                output.append(config)
                counter += 1
                pbar.update(1)
        return output

    def next(self):
        """
        Returns: the next model/config in the list

        """
        if self.verbose:
            print(f"Current configuration: {self.current} of {self.max}")
        config = self.configs[self.current]
        # config = {"regularization": config[0], "learning_rate": config[1], "loss_function": config[2],
        #           "activation_functions": list(config[7]), "output_activation": self.output_activation,
        #           "num_of_units": list(config[6]), "num_layers": config[5], "momentum": config[3],
        #           "optimizer": config[4],
        #           "initializers": config[8], "nesterov": True if config[9] == 'True' else False,
        #           "input_shape": self.input_shapes, "callbacks": self.callbacks, "verbose": self.verbose,
        #           "outputs": self.outputs, "model_name": 'Model' + str(self.current)}

        return config

    def __iter__(self):
        return self

    def __next__(self):
        self.current += 1
        if self.current < self.max:
            return self.next()
        else:
            raise StopIteration

    def len(self):
        return self.max

    @staticmethod
    def units_activations_per_layer_combinations(activation_functions, units, layers):
        if isinstance(activation_functions[0], list) and len(activation_functions[0]) == layers:
            return product(product(units, repeat=layers), activation_functions[0])
        return product(product(units, repeat=layers), product(activation_functions, repeat=layers))

    def get_configs(self):
        return self.configs
        # tmp_configs = []
        # for i, config in enumerate(self.configs):
        #     t_config = {"regularization": config[0], "learning_rate": config[1], "loss_function": config[2],
        #                 "activation_functions": list(config[7]), "output_activation": self.output_activation,
        #                 "num_of_units": list(config[6]), "num_layers": config[5], "momentum": config[3],
        #                 "optimizer": config[4],
        #                 "initializers": config[8], "nesterov": True if config[9] == 'True' else False,
        #                 "input_shape": self.input_shapes, "callbacks": self.callbacks, "verbose": self.verbose,
        #                 "outputs": self.outputs, "model_name": 'Model' + str(i)}
        #     tmp_configs.append(t_config)
        # return tmp_configs



__all__ = ['WandbCallback']

import wandb


class WandbCallback:
    """
    Wandb callback to log metrics
    """
    def __init__(self, run_name, project='exclusiveAI', config=None):
        self.name = 'wandb'
        self.run = ''
        self.config = config
        self.project = None
        self.run = None
        self.log_dict = {}
        self.log_list = []
        wandb.init(project=project, name=run_name, config=config)
        wandb.define_metric("train/step")
        # set all other train/ metrics to use this step
        wandb.define_metric("train/*", step_metric="train/step")
        wandb.define_metric("val/*", step_metric="train/step")

    def __call__(self, model, *args, **kwargs):
        for name in model.history:
            if name.split('_')[0] == 'val':
                self.log_dict['val/' + name.split('_')[-1]] = model.history[name][-1]
            else:
                self.log_dict['train/' + name] = model.history[name][-1]
            self.log_dict['train/step'] = model.curr_epoch
        wandb.log(self.log_dict)

    def reset(self):
        self.log_dict = {}
        self.log_list = []
        wandb.log(self.log_dict)
        self.run = wandb.run
        self.project = wandb.run.project
        self.config = wandb.config
        wandb.define_metric("train/step")
        # set all other train/ metrics to use this step
        wandb.define_metric("train/*", step_metric="train/step")
        wandb.define_metric("val/*", step_metric="train/step")

    def close(self):
        wandb.finish()


from exclusiveAI.components.Layers.Layer import Layer
from exclusiveAI.components.ActivationFunctions import Linear
from exclusiveAI.components.Initializers import *
import numpy as np


class InputLayer(Layer):
    """
    Specialization of the Layer class to represent an input layer.

    Args:
        input_len (int): The length of the input.
        units (int): The number of units in the layer.
    Attributes:
        units (int): The number of units in the layer.
        initializer (Initializer): The initializer to use.
        is_trainable (bool): Whether the layer is trainable or not.
        activation_func (ActivationFunction): The activation function to use.
        input_len (int): The length of the input.
        input (np.ndarray): The input to the layer.
    """
    def __init__(self, input_len, units: int):
        super().__init__(
            units=units,
            initializer=Uniform(low=-1, high=1),
            is_trainable=False,
            activation_func=Linear(),
        )
        self.input_len = input_len
        self.input = None

    def __str__(self) -> str:
        return super().__str__()

    def initialize(self, name: str = '', verbose: bool = False, **kwargs):
        # self.weights = self.initializer.ones(shape=(self.units+1, self.units))
        self.weights = self.initializer.initialize(shape=(self.units + 1, self.units))
        self.name = name
        self.verbose = verbose
        self.is_initialized = True
        if self.verbose:
            print(f"Initializing {self.name}")
            print(f"Input shape: {self.input_len}")
            print(f"Input: {self.input}")
            print(f"Output shape: {self.units}")
            print(f"Activation function: {self.activation_func.name}")
            print(f"Initializer: {self.initializer.name}")
            print(f"Trainable: {self.is_trainable}")
        return self

    def feedforward(self, input):
        if not self.is_initialized:
            raise Exception("Layer not initialized")
        self.input = input
        local_input = input

        #NOTE: this layer simply uses the local input as nets and an identity function as activation function
        self.nets = local_input # calculate the net input for current unit
        self.output = self.activation_func.function(self.nets)
        return self.input

    def backpropagate(self, **kwargs):
        if not self.is_initialized:
            raise Exception("Layer not initialized")

        input = np.insert(self.input, 0, 1, axis=-1)  # adding bias to input
        next_weights = self.next.weights[1:, :].T
        # calculate the product between the error signal and incoming weights from current unit
        self.error = self.next.error @ next_weights
        self.error = self.activation_func.derivative(self.nets) * self.error
        return np.dot(input.T, self.error)



import numpy as np

from exclusiveAI.components.Layers.Layer import Layer
from exclusiveAI.components.ActivationFunctions import ActivationFunction
from exclusiveAI.components.Initializers import Initializer
from exclusiveAI.components.LossFunctions import LossFunction


class OutputLayer(Layer):
    """
    Specialization of the Layer class to represent an output layer.
    Args:

        activation_function: Activation function to use.
        units: Number of units in the layer.
        initializer: Initializer to use.
        loss_function: Loss function to use.

    Attributes:
        initializer:  Initializer to use.
        activation_func: Activation function to use.
        units: Number of units in the layer.
        loss_function: Loss function to use.
    """
    def __init__(self, activation_function: ActivationFunction, units: int, initializer: Initializer,
                 loss_function: LossFunction):
        super().__init__(
            initializer=initializer,
            activation_func=activation_function,
            units=units,
        )
        self.loss_function = loss_function

    def get_output(self):
        return self.output

    def backpropagate(self, y_true: np.ndarray):
        if not self.is_initialized:
            raise Exception("Layer not initialized")

        # check if y_true shape is n,1
        if y_true.shape == (y_true.shape[0],):
            y_true = y_true.reshape(-1, 1)

        # calculate the product between the error signal and incoming weights from current unit
        loss_btw_output_and_y_true = - self.loss_function.function_derivative(y_true, self.output)
        self.error = loss_btw_output_and_y_true * self.activation_func.derivative(self.nets)

        return np.dot(np.insert(self.prev.output, 0, 1, axis=-1).T, self.error)



# import cupy 
from .Device import Device


class cuda(Device):
    def __init__(self):
        self.name = "cuda"
        self.device_id = 0
        # self.device = cupy.cuda.Device(self.device_id)
        self.device.use()
        self.device_properties = self.device.get_attributes()
        self.device_properties["name"] = self.device.name
        self.device_properties["compute_capability"] = self.device.compute_capability
        self.device_properties["total_memory"] = self.device.total_memory
        self.device_properties["memory_bandwidth"] = self.device.memory_bandwidth
        self.device_properties["clock_rate"] = self.device.clock_rate
        self.device_properties["multi_processor_count"] = self.device.multi_processor_count
        self.device_properties["max_threads_per_block"] = self.device.max_threads_per_block
        self.device_properties["max_block_size"] = self.device.max_block_size
        
    # def activationfn_sigmoid(self, v):
        # return cupy.sigmoid(v)


import numpy as np
from tqdm import tqdm
import pandas as pd

__all__ = ['get_best_model']


def get_best_model(results, num_models):
    # Convert the list of tuples to a Pandas DataFrame
    df = pd.DataFrame(results, columns=['score', 'model', 'config'])

    # Order by score, then by std, and finally by model_dim
    df.sort_values(by=['score'], inplace=True)
    df.reset_index(inplace=True)
    # Select the top num_models
    df = df.head(num_models)

    return np.array(df['model']), np.array(df['config'])


__all__ = ['Optimizer']

from exclusiveAI.components.Schedulers import LearningRateScheduler


class Optimizer:
    """
    Optimizer class
    Args:
        learning_rate (float): learning rate
        lr_scheduler (LearningRateScheduler): learning rate scheduler
    Attributes:
        learning_rate (float): learning rate
        lr_scheduler (LearningRateScheduler): learning rate scheduler
        old_dw (list): list of old deltas of weights
    """
    def __init__(self, lr_scheduler=None, **kwargs):
        self.lr_scheduler = lr_scheduler
        self.old_dw = []

    def calulate_deltas(self, model, y_true, x):
        """
        Calculate deltas
        Args:
            model: the current model
            y_true: the target value
            x: the input value

        Returns:
            deltas: the deltas of weights
        """
        model.predict(x)

        layers = list(reversed(model.layers)) #to start from output layer
        output_layer = layers.pop(0)
        deltas = []
        deltas.insert(0, output_layer.backpropagate(y_true))

        for layer in layers:
            deltas.insert(0, layer.backpropagate())
        return deltas

    def update_lr(self):
        self.learning_rate = self.lr_scheduler.update(self.learning_rate)

    def update(self, model, y_true, x):
        pass



import numpy as np
from .ActivationFunction import ActivationFunction

__all__ = ["Sigmoid"]


class Sigmoid(ActivationFunction):
    """
    Sigmoid activation function.
    """

    def __init__(self):
        super().__init__(
            name="Sigmoid",
            function=self.function,
            derivative=self.derivative
        )

    @staticmethod
    def function(x):
        return np.ones(x.shape) / (np.ones(x.shape) + np.exp(-x, dtype='float32'))

    @staticmethod
    def derivative(x):
        return Sigmoid.function(x) * (np.ones(x.shape) - Sigmoid.function(x))



__all__ = ['Layer']

from exclusiveAI.components.ActivationFunctions import ActivationFunction
from exclusiveAI.components.Initializers import Initializer
import numpy as np


class Layer:
    """
    Represents a generic layer of the neural network.

    Args:
        units (int): Number of units in the layer.
        initializer (Initializer): Initializer object to initialize the weights.
        activation_func (ActivationFunction): Activation function object to apply to the nets.
        is_trainable (bool): Whether the layer is trainable or not. Defaults to True.
    Attributes:
        next (Layer): the next Layer of the current one.
        prev (Layer): the previous Layer of the current one.
        name (str): the name of the layer.
        output (np.ndarray): the output of the layer.
        weights (np.ndarray): the connection weights of the layer.
        units (int): the number of units in the layer.
        is_trainable (bool): whether the layer is trainable or not.
        activation_func (ActivationFunction): the activation function object to apply to the nets.
        is_initialized (bool): whether the layer is initialized or not.
        error (np.ndarray): the error of the layer.
        nets (np.ndarray): the nets of the layer.
        verbose (bool): whether to print the information about the layer.
    """
    def __init__(self, units: int, initializer: Initializer, activation_func: ActivationFunction,
                 is_trainable: bool = True) -> object:
        # Layers
        self.next: Layer = None
        self.prev: Layer = None
        self.name = ''
        self.output = None
        self.weights = None
        self.units = units
        self.is_trainable = is_trainable
        self.activation_func = activation_func
        self.initializer = initializer
        self.is_initialized = False
        self.error = None
        self.nets = None
        self.verbose = None
        
    def __str__(self) -> str:
        """
        Print the name of the layer and the number of units in it.
        Returns:
            str: the name of the layer and the number of units in it.
        """
        return f"{self.name} - {self.units}"

    def initialize(self, prev, name: str = '', verbose: bool = False):
        """
        Layer initialization.

        Args:
            prev (Layer): the previous Layer of the current one.
            name (str): the name of the layer.
            verbose (bool): whether to print the information about the layer.

        Returns:
            Layer: the current Layer.

        Raises:
            Exception: if the layer is already initialized.

        Notes:
            The first unit of the layer is the bias unit.
            The weights of the layer are initialized using the `initializer` object.
            The name of the layer is set to `name` if provided, otherwise it is set to the class name.
            The previous layer of the current one is set to `prev`.
            The verbose flag is set to `verbose` if provided, otherwise it is set to False.
            The is_initialized flag is set to True.
            If the verbose flag is True, it prints the information about the layer.
            The weights of the layer are initialized using the `initializer` object.
            The output of the layer is initialized to None.
            The error of the layer is initialized to None.
            The nets of the layer are initialized to None.
        """
        self.weights = self.initializer.initialize(shape=(prev.units + 1, self.units))
        self.name = name
        prev.next = self
        self.prev = prev
        self.verbose = verbose
        self.is_initialized = True
        if self.verbose:
            print(f"Initializing {self.name}")
            print(f"Input shape: {self.prev.units}")
            print(f"Weights shape: {self.weights.shape}")
            print(f"Output shape: {self.units}")
            print(f"Activation function: {self.activation_func.name}")
            print(f"Initializer: {self.initializer.name}")
            print(f"Trainable: {self.is_trainable}")
        return self

    def feedforward(self, input):
        """
        Apply the feedforward step to the input.

        Args:
            input: input to the layer.

        Raises:
            Exception: if the layer is not initialized.

        Returns:
            np.ndarray: the output of the layer (activation function applied to the nets)

        """
        if not self.is_initialized:
            raise Exception("Layer not initialized")

        local_input = np.insert(input, 0, 1, axis=-1)  # adding bias to input

        self.nets = (local_input @ self.weights)  # calculate the net input for current unit
        self.output = self.activation_func.function(self.nets)
        return self.output

    def backpropagate(self, **kwargs):
        """
        Backpropagation step.

        Raises:
            Exception: if the layer is not initialized.

        Returns:

        """
        if not self.is_initialized:
            raise Exception("Layer not initialized")

        previous_output = np.insert(self.prev.output, 0, 1, axis=-1)  # adding bias to input

        next_weights = self.next.weights[1:, :].T

        # calculate the product between the error signal and incoming weights from current unit
        self.error = np.matmul(self.next.error, next_weights, dtype=np.longdouble)
        self.error = self.error * self.activation_func.derivative(self.nets)

        res = np.dot(previous_output.T, self.error)

        return res

    def get_weights(self):
        return self.weights

    def get_error(self):
        return self.error

    def set_weights(self, weights):
        self.weights = weights



from exclusiveAI.components.ActivationFunctions import *
from exclusiveAI.components.LossFunctions import *
from exclusiveAI.components.Initializers import *
from exclusiveAI.components.CallBacks import *
from exclusiveAI.components.Optimizers import *
from exclusiveAI.components.Layers import *
from exclusiveAI.components import NeuralNetwork

InitializersNames = {
    'gaussian': Gaussian,
    'uniform': Uniform,
}

CallbacksNames = {
    'earlystopping': EarlyStoppingCallback,
    'wandb': WandbCallback,
}

LossFunctionsNames = {
    'meansquarederror': MeanSquaredError,
    'meaneuclideandistance': MeanEuclideanDistance,
    'mse': MeanSquaredError,
    'mee': MeanEuclideanDistance,
    'crosscorrelation': CrossCorrelation,
}

ActivationFunctionsNames = {
    'sigmoid': Sigmoid,
    'tanh': Tanh,
    'relu': ReLU,
    'linear': Linear,
    'softmax': Softmax,
}

OptimizersNames = {
    'sgd': SGD,
    'adam': Adam,
}


class Composer:
    """
    Compose the building block to build the model.
    Args:
        regularization (float): The regularization parameter to be used for the model.
        learning_rate (float): The learning rate to be used for the model.
        loss_function (str): The loss function to be used for the model.
        activation_functions (list): The activation functions to be used for the model.
        num_of_units (list): The number of units in each layer.
        num_layers (int): The number of layers in the model.
        momentum (float): The momentum parameter to be used for the model.
        optimizer (str): The optimizer to be used for the model.
        initializers (list): The initializers to be used for the model.
        input_shape (tuple): The input shape of the model.
        callbacks (list): The callbacks to be used for the model.
        verbose (bool): Whether to print out the progress of the model.
        outputs (int): The number of outputs of the model.
        config (dict): The configuration of the model. Note: you can take a config dict in input to build the model instead of each single param.
    Attributes:

    """

    def __init__(self,
                 regularization: float = None,
                 learning_rate: float = None,
                 loss_function: str = None,
                 activation_functions=None,
                 output_activation=None,
                 num_of_units: list = None,
                 model_name: str = 'test',
                 num_layers: int = None,
                 nesterov: bool = False,
                 momentum: float = None,
                 optimizer: str = None,
                 beta1: float = None,
                 beta2: float = None,
                 initializers=None,
                 eps: float = None,
                 input_shape=None,
                 callbacks=None,
                 verbose=False,
                 outputs=1,
                 config: {} = None,
                 ):
        if config is not None:
            if isinstance(config, dict):
                regularization = config.get('regularization', regularization)
                learning_rate = config.get('learning_rate', learning_rate)
                loss_function = config.get('loss_function', loss_function)
                activation_functions = config.get('activation_functions', activation_functions)
                output_activation = config.get('output_activation', output_activation)
                num_of_units = config.get('num_of_units', num_of_units)
                num_layers = config.get('num_layers', num_layers)
                momentum = config.get('momentum', momentum)
                optimizer = config.get('optimizer', optimizer)
                initializers = config.get('initializers', initializers)
                callbacks = config.get('callbacks', callbacks)
                nesterov = config.get('nesterov', nesterov)
                verbose = config.get('verbose', verbose)
                outputs = config.get('outputs', outputs)
                input_shape = config.get('input_shape', input_shape)
                model_name = config.get('model_name', model_name)
            # else:
            #     regularization = config[0]
            #     learning_rate = config[1]
            #     loss_function = config[2]
            #     activation_functions = list(config[7])
            #     output_activation = config
            #     t_config = {"regularization": config[0], "learning_rate": config[1], "loss_function": config[2],
            #                 "activation_functions": list(config[7]), "output_activation": self.output_activation,
            #                 "num_of_units": list(config[6]), "num_layers": config[5], "momentum": config[3],
            #                 "optimizer": config[4],
            #                 "initializers": config[8], "nesterov": True if config[9] == 'True' else False,
            #                 "input_shape": self.input_shapes, "callbacks": self.callbacks, "verbose": self.verbose,
            #                 "outputs": self.outputs, "model_name": 'Model' + str(i)}
        if input_shape is None:
            # Error can't initialize
            raise ValueError("Parameter input_shape can't be None")
        if num_layers is None:
            raise ValueError("Parameter num_layers can't be None")
        if num_of_units is None:
            raise ValueError("Parameter num_of_units can't be None")
        if len(num_of_units) != num_layers:
            raise ValueError("Parameter num_of_units must have the same length as num_layers")
        if optimizer is None:
            optimizer = SGD()
        if initializers is None:
            initializers = [Gaussian()]
        if callbacks is None:
            callbacks = []
        if loss_function is None:
            loss_function = MeanSquaredError()
        if activation_functions is None:
            activation_functions = [Sigmoid()]
        if output_activation is None:
            output_activation = Sigmoid()
        if not isinstance(activation_functions, list):
            activation_functions = [activation_functions]
        if not isinstance(initializers, list):
            initializers = [initializers]
        if not isinstance(callbacks, list):
            callbacks = [callbacks]
        if len(initializers) > 1:
            if len(initializers) != num_layers:
                raise ValueError("Parameter initializers must have the same length as num_layers")
            else:
                self.manyInitializers = True
        else:
            self.manyInitializers = False
        if len(activation_functions) - 1 > 1:
            if len(activation_functions) != num_layers:
                print(activation_functions, num_layers, config)
                raise ValueError("Parameter activation_functions must have the same length as num_layers")
            else:
                self.manyActivations = True
        else:
            self.manyActivations = False

        # Get each initializer for each layer
        # If one, every layer will have the same initializer
        # If more than one, each layer will have its own initializer and the # of initializers must be equal to # of layer
        self.initializers = [InitializersNames[initializer.lower()]() if isinstance(initializer, str) else initializer
                             for initializer in initializers]

        self.callbacks = [self.str_to_callback(callback, model_name)
                          for callback in callbacks]

        self.loss_function = LossFunctionsNames[loss_function.lower()]() \
            if isinstance(loss_function, str) else loss_function

        self.activation_functions = [ActivationFunctionsNames[activation_function.lower()]()
                                     if isinstance(activation_function, str) else activation_function
                                     for activation_function in activation_functions]

        self.output_activation = ActivationFunctionsNames[output_activation.lower()]() if isinstance(output_activation,
                                                                                                     str) else output_activation

        self.optimizer = OptimizersNames[optimizer.lower()](regularization=regularization,
                                                            learning_rate=learning_rate,
                                                            momentum=momentum,
                                                            nesterov=nesterov,
                                                            beta1=beta1,
                                                            beta2=beta2,
                                                            eps=eps
                                                            ) \
            if isinstance(optimizer, str) else optimizer

        self.num_of_units = num_of_units
        self.input_shape = input_shape
        self.num_layers = num_layers
        self.output_units = outputs
        self.verbose = verbose

    def compose(self, regression: bool = False) -> NeuralNetwork.NeuralNetwork:
        """
        Compose the building block to build the model.
        Returns:
            model: The model composed.
        """
        layers = []
        input_layer = InputLayer(self.input_shape[0], self.input_shape[-1])
        layers.append(input_layer)
        for i in range(self.num_layers):
            layers.append(Layer(self.num_of_units[i], self.initializers[i if self.manyInitializers else 0],
                                self.activation_functions[i if self.manyActivations else 0]))
        output_layer = OutputLayer(units=self.output_units, activation_function=self.output_activation,
                                   initializer=self.initializers[-1], loss_function=self.loss_function)
        layers.append(output_layer)

        model = NeuralNetwork.NeuralNetwork(optimizer=self.optimizer,
                                            callbacks=self.callbacks,
                                            metrics=['mse', 'mae', 'mee', 'binary_accuracy'] if not regression else [
                                                'mse', 'mae', 'mee'],
                                            layers=layers,
                                            verbose=self.verbose)

        return model

    @staticmethod
    def str_to_callback(callback, model_name):
        if isinstance(callback, str):
            if '_' in callback:
                call = callback.split('_')
                local_eps = 1e-4
                patience_limit = 50
                restore_weights = False
                penalty = False
                for i, arg in enumerate(call[1:]):
                    if i == 0:
                        local_eps = float(arg)
                    elif i == 1:
                        patience_limit = int(arg)
                    elif i == 2:
                        restore_weights = True if arg.lower() == 'true' else False
                    elif i == 3:
                        penalty = True if arg.lower() == 'true' else False
                return CallbacksNames[call[0].lower()](run_name=model_name, eps=local_eps, penalty=penalty,
                                                       patience_limit=patience_limit, restore_weights=restore_weights)
            else:
                return CallbacksNames[callback.lower()](run_name=model_name)
        else:
            return callback



from .Metric import Metric
import numpy as np

__all__ = ['MAE']


class MAE(Metric):
    """
    Mean Absolute Error (MAE)
    """

    def __init__(self):
        super().__init__(name='mae', f=self.function)

    @staticmethod
    def function(y_true, y_pred):
        return np.mean(np.abs(y_true - y_pred, dtype='float32'), dtype='float32')



import numpy as np

from .Optimizer import Optimizer
from exclusiveAI.components.ActivationFunctions import ActivationFunction


class SGD(Optimizer):
    """
    Stochastic Gradient Descent algorithm
    Args:
        nesterov (bool): nesterov parameter
        momentum (float): momentum parameter
        learning_rate (float): learning rate parameter
        regularization (float): regularization parameter
    Attributes:
        momentum (float): momentum parameter
        learning_rate (float): learning rate parameter
    """
    def __init__(self, nesterov: bool = False, momentum: float = 0.0, learning_rate: float = 0.001, regularization: float = 0.0, **kwargs):
        super().__init__()
        self.learning_rate = learning_rate if learning_rate is not None else 0.001
        self.momentum = momentum if momentum is not None else 0.0
        self.regularization = regularization if regularization is not None else 0.0
        self.nesterov = nesterov if nesterov is not None else False

    def update(self, model, x, y_true):
        """
        The algorithm.
        Args:
            model: current model
            x: input
            y_true: target
        """
        if not self.old_dw:
            dw = self.calulate_deltas(model, y_true, x)
            for layer, delta in zip(model.layers, dw):
                layer.weights = layer.weights + self.learning_rate * delta - layer.weights * self.regularization * 2
            self.old_dw = dw
            return

        new_deltas = []
        
        if self.nesterov:
            for layer, old_delta in zip(model.layers, self.old_dw):
                layer.weights = layer.weights + self.momentum * old_delta
             
        dw = self.calulate_deltas(model, y_true, x)
        
        if self.nesterov:
            for layer, old_delta in zip(model.layers, self.old_dw):
                layer.weights = layer.weights - self.momentum * old_delta

        for layer, delta, old_delta in zip(model.layers, dw, self.old_dw):
            new_delta = self.learning_rate * delta + self.momentum * old_delta
            new_deltas.append(new_delta)
            layer.weights = layer.weights + new_delta - layer.weights * self.regularization * 2

        self.old_dw = new_deltas



import numpy as np
from .ActivationFunction import ActivationFunction

__all__ = ["Linear"]


class Linear(ActivationFunction):
    """
    Linear activation function.
    """
    def __init__(self):
        super().__init__(
            name="Linear",
            function=self.function,
            derivative=self.derivative,
        )

    @staticmethod
    def function(x):
        return x

    @staticmethod
    def derivative(x):
        return np.ones(shape=x.shape)


import numpy as np
from .ActivationFunction import ActivationFunction

__all__ = ["ReLU"]


class ReLU(ActivationFunction):
    """
    ReLU activation function.
    """
    def __init__(self) -> None:
        super().__init__(
            name="ReLU",
            function=self.function,
            derivative=self.derivative,
        )

    @staticmethod
    def function(x):
        return np.maximum(x, np.zeros(shape=x.shape))

    @staticmethod
    def derivative(x):
        return np.where(x < 0, np.zeros(shape=x.shape), np.ones(shape=x.shape))



from .Metric import Metric
import numpy as np

__all__ = ['MSE']


class MSE(Metric):
    """
    Mean squared error (MSE)
    """
    def __init__(self):
        super().__init__(name='mse', f=self.function)

    @staticmethod
    def function(y_pred, y_true):
        return np.mean(np.sum(np.square(y_pred - y_true, dtype='float32'), axis=1, dtype='float32'), dtype='float32')



import numpy as np
from .LossFunction import LossFunction

__all__ = ["MeanEuclideanDistance"]


class MeanEuclideanDistance(LossFunction):
    """
    Mean Euclidean Distance loss fucntion.
    """
    def __init__(self) -> None:
        super().__init__(
            name="Mean Euclidean Distance",
            function=self.function,
            derivative=self.derivative
        )

    @staticmethod
    def function(y_true, y_pred):
        return np.mean(np.sqrt(np.sum((np.square(y_true-y_pred, dtype='float32')), axis=1, dtype='float32'), dtype='float32'), dtype='float32')

    @staticmethod
    def derivative(y_true, y_pred):
        return -1 / y_pred.shape[0] * (y_true-y_pred) / np.linalg.norm(y_true - y_pred, axis=1).reshape(-1, 1)



__all__ = ["ActivationFunction"]


class ActivationFunction:
    """
    Class that represents an activation function.
    Attributes:
        name (str): The name of the activation function.
        function (function): The function itself.
        derivative (function): The activation function derivative.
    """
    def __init__(self, name, function, derivative):
        self.name = name
        self.function = function
        self.derivative = derivative

    def __str__(self) -> str:
        return self.name



from .Metric import Metric
import numpy as np


class BinaryAccuracy(Metric):
    """
		Binary Accuracy
	"""

    def __init__(self):
        super().__init__(name="binary_accuracy",
                         f=self.function)

    @staticmethod
    def function(y_true, y_pred):
        return np.average((np.round(np.abs(y_true)) == np.round(np.abs(y_pred))))



__all__ = ["Metric"]

class Metric:
    """
    A class that represents a metric
    Args:
        name (str): The name of the metric
        f (function): The function that calculates the metric
    Attributes:
        name (str): The name of the metric
        f (function): The function that calculates the metric
    """
    def __init__(self, name, f):
        self.name = name
        self.f = f

    def __call__(self, *args, **kwargs):
        return self.f(*args, **kwargs)





from exclusiveAI.components.Layers import Layer
from .Optimizer import Optimizer
import numpy as np


class Adam(Optimizer):
    def __init__(self, beta1: float = 0.9, beta2: float = 0.999, eps: float = 1e-8, learning_rate: float = 0.001, **kwargs):
        super().__init__()
        self.learning_rate = learning_rate if learning_rate is not None else 0.001
        self.mean = 0.0
        self.variance = 0.0
        self.beta1 = beta1 if beta1 is not None else 0.9
        self.beta2 = beta2 if beta2 is not None else 0.999
        self.eps = eps if eps is not None else 1e-8

    def update(self, model, x, y_true):
        
        deltas = self.calulate_deltas(model, y_true, x)

        if self.mean == 0.0:
            self.mean = [np.zeros_like(delta) for delta in deltas]
            self.variance = self.mean
            
        new_mean = []
        new_var = []
        for meani, variancei, deltai in zip(self.mean, self.variance, deltas):
            new_mean.append(self.beta1 * meani + (1 - self.beta1) * deltai)
            new_var.append(self.beta2 * variancei + (1 - self.beta2) * deltai ** 2)
        new_mean = [meani / (1 - self.beta1) for meani in new_mean]
        new_var = [variancei / (1 - self.beta2) for variancei in new_var]
        self.mean = new_mean
        self.variance = new_var

        for layer, mean, var in zip(model.layers, new_mean, new_var):
            layer.weights = layer.weights - (self.learning_rate * mean / (var ** 0.5 + self.eps))


__all__ = ['utils', 'train_split', 'one_hot_encoding', 'plot_history']

import numpy as np
import matplotlib.pyplot as plt


def split_batches(inputs, input_label, batch_size):
    """
    Split the inputs into batches of size batch_size.
    """
    for i in range(0, len(inputs), batch_size):
        # if last
        if i + batch_size > len(inputs):
            batch_size = len(inputs) - i
        yield inputs[i:i + batch_size], input_label[i:i + batch_size]


def myzip(*iterables):
    sentinel = object()
    iterators = [iter(it) for it in iterables if it]
    while iterators:
        result = []
        for it in iterators:
            elem = next(it, sentinel)
            if elem is sentinel:
                return
            result.append(elem)
        yield tuple(result)


def train_split(inputs, input_label, split_size=0.2, shuffle=True, random_state=42):
    """
    Split the data into training and test sets.
    """
    indices = np.arange(len(inputs))
    if shuffle:
        np.random.seed(random_state)
        np.random.shuffle(indices)
    # Left size is training set, 1- split size * the # of the inputs
    left_indices = indices[:int((1 - split_size) * len(indices))]
    # Right size is the test/validation set, inputs - training set
    right_indices = indices[len(left_indices):]
    return inputs[left_indices], input_label[left_indices], inputs[right_indices], input_label[
        right_indices], left_indices, right_indices


def plot_history(lines: dict, fig_size=(10, 6)):
    plt.figure(figsize=fig_size)
    for elem in lines:
        plt.plot(lines[elem], label=elem)
    plt.legend()
    plt.show()


def one_hot_encoding(y):
    """
    One hot encoding.
    """
    for column in y.T:
        col_max = column.max()
        new_col = np.zeros((len(column), col_max))
        for i in range(len(column)):
            new_col[i][column[i] - 1] = 1
        column = new_col
        y = np.hstack((y, column))
        y = y[:, 1:]
    return y





__all__ = ["Softmax"]

import numpy as np
from .ActivationFunction import ActivationFunction


class Softmax(ActivationFunction):
    """
    Softmax activation function.
    """
    def __init__(self):
        super().__init__(
            name="Softmax",
            function=self.function,
            derivative=self.derivative,
        )

    @staticmethod
    def function(x):
        return np.exp(x) / np.sum(np.exp(x), axis=0)

    @staticmethod
    def derivative(x):
        return np.exp(x - x.max()) / np.sum(np.exp(x - x.max()), axis=0)


import pickle

import numpy as np
from exclusiveAI.components.Optimizers import Optimizer
from exclusiveAI.components.Metrics import MetricUtils
from .exclusiveAIModel import exclusiveAIModel
from exclusiveAI import utils
from tqdm import tqdm


class NeuralNetwork(exclusiveAIModel):
    """
    This class is used to create a neural network model.

    Args:

        layers (list): A list of layers to be added to the model.
        optimizer (Optimizer): The optimizer to be used for training.
        callbacks (list): A list of callbacks to be used during training.
        metrics (list): A list of metrics to be used during training.
        verbose (bool): Whether to print out the progress of the model.
        shuffle (bool): Whether to shuffle the data before training.
    Attributes:

        optimizer (Optimizer): The optimizer to be used for training.
        callbacks (list): A list of callbacks to be used during training.
        layers (list): A list of layers to be added to the model.
        verbose (bool): Whether to print out the progress of the model.
        early_stop (bool): Whether to use early stopping as stopping criteria.
        name (str): The name of the model.
        curr_epoch (int): The current epoch of the model.
        metrics (list): A list of metrics to be used during training.
        history (dict): A dictionary containing the training history metrics.
    """
    def __init__(self,
                    layers: list,
                    optimizer: Optimizer,
                    callbacks: list,
                    metrics: [],
                    verbose: bool = False,
                    shuffle: bool = True
                    ):
        super().__init__(layers, optimizer, callbacks, metrics, verbose, shuffle)

    def train(self,
              inputs: np.array,
              input_label: np.array,
              val: np.array = None,
              val_labels: np.array = None,
              epochs=100, batch_size=32, name: str = '', disable_line=False):
        """
        Perform the model training on input data and label.

        Args:
            inputs (np.array): the input data
            input_label (np.array) the label data
            val (np.array): the validation data
            val_labels (no.array): the validation data label
            epochs (int): the number of epochs
            batch_size (int): the size of a batch used in minibatch approach
            disable_line (bool): show line
            name: name of the model

        Returns: training history

        """

        # cast array to float64
        # check if both val and val_label are provided
        if val is not None and val_labels is not None:
            # check if val and val_label have the same shape
            if val.shape[0] != val_labels.shape[0]:
                raise ValueError("val and val_label must have the same shape")
        MetricUtils.initialize_history(self, val is not None)
        for callback in self.callbacks:
            callback.reset()
        with tqdm(total=epochs, desc="Epochs", colour="white", disable=disable_line) as pbar:
            for epoch in range(epochs):
                output = self.predict(inputs)

                val_output = None
                if val is not None:
                    val_output = self.predict(val)

                MetricUtils.add_to_history(self, output, input_label, val_output, val_labels)
                for callback in self.callbacks:
                    callback(self)

                if self.early_stop:
                    [callback.close() for callback in self.callbacks]
                    break

                self.curr_epoch += 1

                if self.shuffle:
                    perm = np.random.permutation(inputs.shape[0])
                    inputs = np.take(inputs, perm, axis=0)
                    input_label = np.take(input_label, perm, axis=0)
                batches = utils.split_batches(inputs, input_label, batch_size)
                for (batch, batch_label) in batches:
                    self.optimizer.update(self, batch, batch_label)
                pbar.update(1)
                if name != '':
                    pbar.set_description(name)
                pbar.set_postfix(self.get_last())

            pbar.close()
        return self.history


__all__ = ['LossFunction']


class LossFunction:
    """
    Loss function class
    Args:

        name: Name of the loss function
        function: Loss function
        derivative: Derivative of the loss function
    """
    def __init__(self, name, function, derivative):
        self.name = name
        self.function = function
        self.function_derivative = derivative

    def __str__(self) -> str:
        return self.name



from .Initializer import Initializer
import numpy as np

__all__ = ['Gaussian']


class Gaussian(Initializer):
    """
    Initialize weights from a random Gaussian distribution.
    Args:
        mean (float): mean of the Gaussian distribution.
        std (float): standard deviation of the Gaussian distribution.
    Attributes:
        name (str): name of the initializer.
        mean (float): mean of the Gaussian distribution.
        std (float): standard deviation of the Gaussian distribution.
    """
    def __init__(self, mean=0, std=0.05):
        super().__init__(name='Gaussian')
        self.mean = mean
        self.std = std

    def initialize(self, shape):
        """
        Args:
            shape (tuple): shape of the weights vector.
        Returns:
            object: returns an array of shape `shape` with values drawn from a random Gaussian distribution.
        """
        return np.random.normal(self.mean, self.std, shape)



from exclusiveAI.utils import train_split
from exclusiveAI.Composer import Composer
from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from .ValidationUtils import get_best_model
from joblib import Parallel, delayed
import numpy as np
from functools import partial

__all__ = ['parallel_hold_out', "hold_out"]

"""
Hold-out
Args:
    models (ConfiguratorGen): a set of models
    input (np.ndarray): input data
    target (np.ndarray): target data
    split_size (float): split size
    shuffle (bool):  true to shuffle data
    seed (int): seed for the random shuffling
    assessment (bool): true to perform model assessment
    debug (bool): true to print debug information
Attributes:
    best_model (neural_network): the best model found
    best_config (dict): the best configuration found
    models (ConfiguratorGen): a set of models
    input (np.ndarray): input data
    target (np.ndarray): target data
    split_size (float): split size
    shuffle (bool):  true to shuffle data
    seed (int): seed for the random shuffling
    assessment (bool): true to perform model assessment
    debug (bool): true to print debug information
"""


def split(training, training_target, split_size=0.2, shuffle=True, seed=42):
    """
    Split the data into TR and VL/TS
    Returns: TR and VL/TS splits with their target values sets

    """
    train, train_target, validation, validation_target, _, _ = train_split(inputs=training,
                                                                           input_label=training_target,
                                                                           split_size=split_size,
                                                                           shuffle=shuffle,
                                                                           random_state=seed)
    return train, train_target, validation, validation_target


def evaluate_model(config, train, train_target, validation, validation_target, assessment, disable_line, metric,
                   batch_size,
                   epochs, regression, number_of_initializations):
    score = 0
    model = None

    for i in range(number_of_initializations):
        tmp_model = Composer(config=config).compose(regression)
        tmp_model.train(train.copy(), train_target.copy(), None if assessment else validation.copy(),
                        None if assessment else validation_target.copy(), disable_line=disable_line,
                        epochs=epochs, batch_size=batch_size)
        if model is None or np.min(model.history[metric]) > np.min(tmp_model.history[metric]):
            model = tmp_model
        score += np.min(tmp_model.history[metric])
    score /= number_of_initializations

    return score, model, config


def parallel_hold_out(configs, training, training_target, metric=None, num_models=1, regression=False,
                      number_of_initializations=1, epochs=100, batch_size=32, return_models_history=False,
                      disable_line=True, workers: any = None, assessment=False):
    """
    The hold out algorithm
    Args:
        configs (list): List of configurations for models to be evaluated
        training: training data
        training_target: training target
        metric: metric to use (e.g, mse)
        epochs: number of epochs
        regression (bool): whether to use regression or classification
        number_of_initializations (int): number of initializations for each model
        batch_size: batch size
        disable_line: whether to disable the line of model training output
        assessment: whether to assess the model or make model selection
        num_models: number of models to be returned if all models
        workers: number of workers to use for parallel
        return_models_history: whether to return the model history or not

    Returns: best configuration for the validation task or model assessment for the test task.

    """
    metric = 'val_mse' if not assessment else 'mse' if metric is None else metric
    train, train_target, validation, validation_target = split(training, training_target)

    evaluate_partial = partial(evaluate_model, train=train.copy(), train_target=train_target.copy(),
                               validation=None if assessment else validation.copy(),
                               validation_target=None if assessment else validation_target.copy(),
                               assessment=assessment, disable_line=disable_line, metric=metric, batch_size=batch_size,
                               epochs=epochs, regression=regression,
                               number_of_initializations=number_of_initializations)

    evaluator = ProcessPoolExecutor(max_workers=workers) if workers is None or workers != 1 else None
    results_func = evaluator.map if evaluator else map

    results = list(tqdm(results_func(evaluate_partial, configs), total=len(configs), desc="Models", colour="white"))

    # results = Parallel(n_jobs=workers)(
    #     delayed(evaluate_model)(config, train.copy(), train_target.copy(),
    #                             validation=None if assessment else validation.copy(),
    #                             validation_target=None if assessment else validation_target.copy(),
    #                             assessment=assessment,
    #                             disable_line=disable_line, metric=metric, batch_size=batch_size,
    #                             epochs=epochs, regression=regression,
    #                             number_of_initializations=number_of_initializations)
    #     for config in tqdm(configs, desc="Models", colour="white")
    # )
    best_models, best_configs = get_best_model(results, num_models)

    if return_models_history:
        if num_models > 1:
            return [model.history for model in best_models], best_configs if not assessment else [
                model.evaluate(validation, validation_target)[0] for model in best_models]
        return best_models[0].history, best_configs[0] if not assessment else \
            best_models[0].evaluate(validation, validation_target)[0]

    if num_models > 1:
        return [model.evaluate(validation, validation_target)[0] for model in
                best_models] if assessment else best_configs
    return best_models[0].evaluate(validation, validation_target)[0] if assessment else best_configs[0]


def hold_out(configs, training, training_target, metric: str = None, epochs=100, regression=False,
             number_of_initializations=1, batch_size=32, disable_line=True, assessment=False,
             return_models_history=False, num_models=1):
    return parallel_hold_out(configs=configs, training=training, training_target=training_target, metric=metric,
                             epochs=epochs, regression=regression, number_of_initializations=number_of_initializations,
                             batch_size=batch_size, disable_line=disable_line, assessment=assessment,
                             return_models_history=return_models_history, num_models=num_models, workers=None)



__all__ = ['Initializer']

import numpy as np


class Initializer:
    """
    Base class for all initializer.
    Attributes:
        name (str): Name of the initializer.
    """
    def __init__(self, name=None):
        self.name = 'Initializer'

    def initialize(self, shape):
        pass



from .Metric import Metric
import numpy as np

__all__ = ['MEE']


class MEE(Metric):
    """
    Mean Euclidean Error (MEE)
    """
    def __init__(self):
        super().__init__(name='mee',
                         f=self.function
                         )

    @staticmethod
    def function(y_true, y_pred):
        return np.mean((np.linalg.norm(y_true - y_pred, axis=1)), dtype='float32')



from exclusiveAI.components import ActivationFunctions
from exclusiveAI.components import LossFunctions
from exclusiveAI.components import Optimizers
from exclusiveAI.components import Initializers
from exclusiveAI.components import NeuralNetwork
from exclusiveAI.components import Layers
from exclusiveAI.components import CallBacks
from exclusiveAI.components import Metrics
from exclusiveAI import utils



__all__ = ['EarlyStoppingCallback']

from ..NeuralNetwork import NeuralNetwork


class EarlyStoppingCallback:
    """
    Implements early stopping
    """

    def __init__(self, eps: float = 1e-4, patience_limit: int = 50, restore_weights: bool = False, penalty=False,
                 metric: str = 'val_mse', **kwargs):
        self.restore_weights = restore_weights
        self.patience_limit = patience_limit
        self.penalty = penalty
        self.metric = metric
        self.stop = False
        self.eps = eps

    def __call__(self, model: NeuralNetwork):
        if self.metric == 'val_mse':
            # check if val_mse in history
            if 'val_mse' not in model.history:
                self.metric = 'mse'
        loss = model.history[self.metric][-1]
        if model.curr_epoch == 0:
            model.best_loss = loss
            model.best_epoch = 0
            model.best_weights = model.get_weights()
            model.patience = 0
            return
        epoch = model.curr_epoch
        if model.best_loss - loss > self.eps:
            model.best_loss = loss
            model.best_epoch = epoch
            model.patience = 0
            model.best_weights = model.get_weights()
        else:
            model.patience += 1
            if self.penalty and model.best_loss - loss < 0:
                model.patience += 1

        if model.patience > self.patience_limit:
            self.stop = True
            model.early_stop = True
            if self.restore_weights:
                model.set_weights(model.best_weights)
                model.history = {key: value[:model.best_epoch] for key, value in model.history.items()}

    def reset(self):
        self.stop = False
        self.metric = 'val_mse'

    def close(self):
        self.reset()



import pickle

import numpy as np
from exclusiveAI.components.Optimizers import Optimizer
from exclusiveAI.components.Metrics import MetricUtils
from exclusiveAI import utils
from tqdm import tqdm


class exclusiveAIModel:
    """
    This class is used to create a neural network model.

    Args:

        layers (list): A list of layers to be added to the model.
        optimizer (Optimizer): The optimizer to be used for training.
        callbacks (list): A list of callbacks to be used during training.
        metrics (list): A list of metrics to be used during training.
        verbose (bool): Whether to print out the progress of the model.
        shuffle (bool): Whether to shuffle the data before training.
    Attributes:

        optimizer (Optimizer): The optimizer to be used for training.
        callbacks (list): A list of callbacks to be used during training.
        layers (list): A list of layers to be added to the model.
        verbose (bool): Whether to print out the progress of the model.
        early_stop (bool): Whether to use early stopping as stopping criteria.
        name (str): The name of the model.
        curr_epoch (int): The current epoch of the model.
        metrics (list): A list of metrics to be used during training.
        history (dict): A dictionary containing the training history metrics.
    """

    def __init__(self,
                 layers: list,
                 optimizer: Optimizer,
                 callbacks: list,
                 metrics: [],
                 verbose: bool = False,
                 shuffle: bool = True
                 ):
        self.optimizer = optimizer
        self.callbacks = callbacks
        self.layers = layers
        self.verbose = verbose
        self.early_stop = False
        self.name = 'Model ' + str(len(self.layers))
        self.curr_epoch = 0
        self.metrics = metrics
        self.history = {}
        self.shuffle = shuffle

        self.best_loss = float('inf')
        self.best_weights = None
        self.best_epoch = 0
        self.patience = 0

        self.initialize()

    def initialize(self):
        """
        Initialize the layers of the model
        """
        self.layers[0].initialize(name='Input', verbose=self.verbose)
        for i, layer in enumerate(self.layers[1:]):
            layer.initialize(self.layers[i], name=('Layer' + str(i)), verbose=self.verbose)

    def train(self, *args):
        pass


    def get_last(self, index=-1):
        """
        Get the last element of the history.
        Returns: the last element of the history.
        """

        return {name: self.history[name][index] for name in self.history}

    def predict(self, input: np.array):
        """
        Apply the feedforward across layers to the input data.
        Args:
            input:

        Returns:

        """
        input = input
        output = None
        for layer in self.layers:
            output = layer.feedforward(input)
            input = output
        return output

    def evaluate(self, input: np.array, input_label: np.array, metrics=None):
        """
        Apply the predict and calculate the metrics on prediction.
        Args:
            input: input data
            input_label: input label
            metrics: metrics to calculate

        Returns: metrics on prediction
        """
        if metrics is None:
            metrics = ['mse', 'binary_accuracy']
        output = self.predict(input)
        return [MetricUtils.calculate(metric, target=input_label, predicted=output) for metric in metrics]

    def get_weights(self):
        weights = []
        for layer in self.layers:
            weights.append(layer.get_weights())
        return weights

    def set_weights(self, weights: list):
        for layer, weight in zip(self.layers, weights):
            layer.set_weights(weight)

    def save(self, filename: str):
        with open(filename, 'wb') as f:
            pickle.dump(self, f)

    @staticmethod
    def load(filename: str):
        with open(filename, 'rb') as f:
            return pickle.load(f)


from setuptools import find_packages, setup

README = open("README.md").read()

setup(
    name='exclusiveAI',
    packages=find_packages(),
    version='0.0.2',
    long_description_content_type="text/markdown",
    url="https://github.com/exclusiveAI/MLProject",
    long_description=README,
    author_email='chuckpaul98@icloud.com',
    install_requires=['numpy', 'matplotlib', 'pandas', 'tqdm', 'joblib', 'wandb'],
    description='An inefficient (for the moment) neural network with numpy',
    setup_requires=['pytest-runner'],
    author='Paul Magos & Francesco Paolo Liuzzi',
    python_requires='>=3.9',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)


import numpy as np
from .ActivationFunction import ActivationFunction

__all__ = ["Tanh"]


class Tanh(ActivationFunction):
    """
    Tanh activation function.
    """
    def __init__(self):
        super().__init__(
            name="Tanh",
            function=self.function,
            derivative=self.derivative,
        )

    @staticmethod
    def function(x):
        return (np.exp(x) - np.exp(-x)) / (np.exp(x) + np.exp(-x))

    @staticmethod
    def derivative(x):
        return np.ones(shape=x.shape) - np.square(Tanh.function(x))



from .LossFunction import LossFunction
import numpy as np

__all__ = ["CrossCorrelation"]


class CrossCorrelation(LossFunction):
    def __init__(self) -> None:
        super().__init__(
            name="Cross Correlation",
            function=self.function,
            derivative=self.derivative,
        )

    @staticmethod
    def function(x: np.ndarray, y: np.ndarray):
        return -1 * np.sum(x * y) / np.sqrt(np.sum(x * x) * np.sum(y * y), dtype=np.float64)

    @staticmethod
    def derivative(x: np.ndarray, y: np.ndarray):
        return (x * y) / np.sqrt(np.sum(x * x) * np.sum(y * y), dtype=np.float64)


import numpy as np
from .LossFunction import LossFunction

__all__ = ["MeanSquaredError"]


class MeanSquaredError(LossFunction):
    """
    Mean Squared Error loss function.
    """
    def __init__(self):
        super().__init__(
            name="Mean Squared Error",
            function=self.function,
            derivative=self.derivative
        )

    @staticmethod
    def function(y_true, y_pred):
        return np.mean(np.sum(np.square(y_true-y_pred, dtype='float32'), axis=1, dtype='float32'), dtype='float32')

    @staticmethod
    def derivative(y_true, y_pred):
        return -(2 / y_true.shape[0]) * (y_true - y_pred)



import pandas as pd
import numpy as np

__all__ = [
    "read_monk",
    "read_monk1",
    "read_monk2",
    "read_monk3"
]


def read_monk(index: int, path="exclusiveAI/datasets/") -> tuple:
    train_set_filepath = f"{path}monks-{index}.train"
    test_set_filepath = f"{path}monks-{index}.test"

    # reading csvs
    train_set_df = pd.read_csv(
        train_set_filepath,
        sep=" ",
        names=["class", "a1", "a2", "a3", "a4", "a5", "a6", "id"]
    ).set_index("id")

    train_label = np.array(train_set_df.pop("class"))
    train_set_df = np.array(train_set_df)

    test_set_df = pd.read_csv(
        test_set_filepath,
        sep=" ",
        names=["class", "a1", "a2", "a3", "a4", "a5", "a6", "id"]
    ).set_index("id")

    test_label = np.array(test_set_df.pop("class"))
    test_set_df = np.array(test_set_df)

    return train_set_df, train_label, test_set_df, test_label


def read_monk1(path="exclusiveAI/datasets/") -> tuple:
    return read_monk(1, path=path)


def read_monk2(path="exclusiveAI/datasets/") -> tuple:
    return read_monk(2, path=path)


def read_monk3(path="exclusiveAI/datasets/") -> tuple:
    return read_monk(3, path=path)



from . import MAE
from . import MEE
from . import MSE
from . import BinaryAccuracy

__all__ = ["string_to_metric", "initialize_history", "add_to_history", "print_history", "calculate"]

MATCH = {
    "mse": MSE,
    "mae": MAE,
    "mee": MEE,
    "binary_accuracy": BinaryAccuracy,
}


def string_to_metric(name):
    """
    Args:
        name (str): name of the metric.
    Raises:
        ValueError: the input string doesn't match any available metric.
    Returns:
        object (Metric): returns a metric matching the input string.
    """
    if name.lower() in MATCH:
        return MATCH[name.lower()]()
    else:
        # Error
        raise ValueError("Unknown metric name: " + name)


def initialize_history(model, val: bool):
    """
    Initializes the metric history of a given model.
    Args:
        model: the model for which to initialize the history.
        val: true if you want to add validation metrics too.
    """
    model.metrics = [string_to_metric(metric) if isinstance(metric, str) else metric for metric in model.metrics]
    model.history = {}
    for metric in model.metrics:
        model.history[metric.name] = []
        if val:
            model.history["val_" + metric.name] = []


def add_to_history(model, y_train_pred, y_train_true, y_val_pred, y_val_true):
    """
    Calculates and adds to the history of a given model the metrics calculated on the training and validation (if specified) data.
    Args:
        model: the model for which to add the metric to its history
        y_train_pred: predicted value by the model for TR
        y_train_true: corresponding target value
        y_val_pred: predicted value by the model for VL
        y_val_true: corresponding target value
    """
    for metric in model.metrics:
        if y_train_true.shape == (y_train_true.shape[0],):
            y_train_true = y_train_true.reshape(-1, 1)
        model.history[metric.name].append(metric(y_train_true, y_train_pred))
        if y_val_pred is not None and y_val_true is not None:
            if y_val_true.shape == (y_val_true.shape[0],):
                y_val_true = y_val_true.reshape(-1, 1)
            model.history['val_' + metric.name].append(metric(y_val_true, y_val_pred))


def print_history(model, val: bool):
    """
    Prints the history of a given model
    Args:
        model: the given model for which to print the metrics history
        val: true if you want to print the validation metrics too
    """
    for metric in model.history:
        print(metric, model.history[metric])
        if val:
            print("val_" + metric, model.history["val_" + metric])


def calculate(func, target, predicted):
    """
    Calculates the metrics specified by the input string.
    Args:
        func (str): the metric name
        target: target value
        predicted: predicted value by the model

    Raises:
        ValueError: if the metric name is unknown

    """
    func = func.lower()
    if func not in MATCH:
        raise ValueError("Unknown metric name: " + func)
    func = MATCH[func]()
    if target.shape == (target.shape[0],):
        target = target.reshape(-1, 1)
    return func(target, predicted)
